#include "stdafx.h"
#include "FrameOperation.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include "vclipDoc.h"

#include <utility>

// FrameOperation

FrameOperation::FrameOperation(CVclipDoc& owner, LONG targetFrame)
: m_pOwner(&owner), m_targetFrame(targetFrame)
{
}

FrameOperation::~FrameOperation()
{
}

void FrameOperation::onModified() const
{
	m_pOwner->SetModifiedFlag();
	m_pOwner->UpdateAllViews(0, m_targetFrame);
}

void FrameOperation::setTargetFrame(LONG targetFrame)
{
	m_targetFrame = targetFrame;
}

LONG FrameOperation::targetFrame() const
{
	return m_targetFrame;
}

// PatchOperation

PatchOperation::PatchOperation(CVclipDoc& owner, LONG targetFrame /*= 0*/, 
															 int targetX /*= 0*/, int targetY /*= 0*/,
															 int width /*= 0*/, int height /*= 0*/)
														 : FrameOperation(owner, targetFrame), 
															 m_targetOrigin(targetX, targetY),
															 m_size(width, height)
{
}

CPoint PatchOperation::targetOrigin() const
{
	return m_targetOrigin;
}

CRect PatchOperation::targetRect() const
{
	return CRect(m_targetOrigin, m_size);
}

int PatchOperation::targetX() const
{
	return m_targetOrigin.x;
}

int PatchOperation::targetY() const
{
	return m_targetOrigin.y;
}

void PatchOperation::resize(CSize newSize)
{
	m_size = newSize;
	onModified();
}

CSize PatchOperation::size() const
{
	return m_size;
}

int PatchOperation::width() const
{
	return m_size.cx;
}

int PatchOperation::height() const
{
	return m_size.cy;
}

void PatchOperation::moveTo(CPoint newTargetOrigin)
{
	m_targetOrigin = newTargetOrigin;
	onModified();
}

// PFrameOperation

void PFrameOperation::accept(IOperationVisitor& visitor) const
{
	if (get()) get()->accept(visitor);
}


bool byTargetFrame::operator()(const PFrameOperation& lhs, const PFrameOperation& rhs) const
{
	return lhs->targetFrame() < rhs->targetFrame();
}

bool byTargetFrame::operator()(const PFrameOperation& lhs, LONG rhs) const
{
	return lhs->targetFrame() < rhs;
}

bool byTargetFrame::operator()(LONG lhs, const PFrameOperation& rhs) const
{
	return lhs < rhs->targetFrame();
}


std::ostream& operator<<(std::ostream& stream, const PFrameOperation& op)
{
	the<FrameOperationFactory>().write(stream, op);
	return stream;
}

// FrameOperationFactory

FrameOperationFactory::FrameOperationFactory()
{
}

/*static*/ FrameOperationFactory& FrameOperationFactory::instance()
{
	static FrameOperationFactory theInstance;
	return theInstance;
}

void FrameOperationFactory::insert(const std::string& typeCode, const ICreateFrameOperation& creator)
{
	m_creators.insert(std::make_pair(typeCode, &creator));
}

PFrameOperation FrameOperationFactory::create(const std::string& typeCode, CVclipDoc& owner) const
{
	Creators::const_iterator it = m_creators.find(typeCode);
	if (it == m_creators.end()) return PFrameOperation();
	return (*it->second)(owner);
}

PFrameOperation FrameOperationFactory::read(std::istream& stream, CVclipDoc& owner) const
{
	if (!stream) return PFrameOperation();

	std::string typeCode;
	stream >> typeCode;
	if (!stream) return PFrameOperation();

	PFrameOperation result = create(typeCode, owner);
	if (result) result->readFrom(stream);

	return result;
}

void FrameOperationFactory::write(std::ostream& stream, PFrameOperation pfo) const
{
	pfo->writeTo(stream << pfo->typeCode() << " ");
}

PFrameOperation FrameOperationFactory::createFramePatch(CVclipDoc& owner, 
	LONG sourceFrame, int sourceX, int sourceY, int sourceWidth, int sourceHeight,
	LONG targetFrame, int targetX, int targetY, HOverlay mode) const
{
	return PFrameOperation(new FramePatch(owner,
		sourceFrame, sourceX, sourceY, sourceWidth, sourceHeight,
		targetFrame, targetX, targetY, mode));
}

PFrameOperation FrameOperationFactory::createFieldPatch(CVclipDoc& owner, 
	LONG sourceFrame, LONG sourceOrder, int sourceX, int sourceY, int sourceWidth, int sourceHeight,
	LONG targetFrame, LONG targetOrder, int targetX, int targetY, HOverlay mode) const
{
	return PFrameOperation(new FieldPatch(owner,
		sourceFrame, sourceOrder, sourceX, sourceY, sourceWidth, sourceHeight,
		targetFrame, targetOrder, targetX, targetY, mode));
}

PFrameOperation FrameOperationFactory::createFrameImage(CVclipDoc& owner,
	LONG targetFrame, const std::string& fileName) const
{
	return PFrameOperation(new FrameImage(owner, targetFrame, fileName));
}

PFrameOperation FrameOperationFactory::createFieldRebuild(CVclipDoc& owner, 
	LONG targetFrame) const
{
	return PFrameOperation(new FieldRebuild(owner, targetFrame));
}

PFrameOperation FrameOperationFactory::createMotionInterpolate(CVclipDoc& owner, 
	LONG targetFrame) const
{
	return PFrameOperation(new MotionInterpolate(owner, targetFrame));
}
