#pragma once

#include "refcount.h"
#include "OverlayMode.h"

#include <windows.h>
#include <istream>
#include <ostream>
#include <map>
#include <string>

class CVclipDoc;
class IOperationVisitor;

// Base class

class FrameOperation : public ref_counted
{
private: // noncopyable
	FrameOperation(const FrameOperation& other);
	FrameOperation& operator=(const FrameOperation& other);

protected:
	virtual ~FrameOperation();

protected: // construction
	explicit FrameOperation(CVclipDoc& owner, LONG targetFrame = 0);

private: // attributes
	CVclipDoc* m_pOwner;
	LONG m_targetFrame;
protected:
	void onModified() const;
	void setTargetFrame(LONG targetFrame);
public:
	LONG targetFrame() const;

	virtual bool usesFrame(LONG frame) const = 0;

public: // serialization
	virtual const char* typeCode() const = 0;
	virtual void writeTo(std::ostream& stream) const = 0;
	virtual void readFrom(std::istream& stream) = 0;

public: // everything else
	virtual void accept(IOperationVisitor& visitor) = 0;
};

class PatchOperation : public FrameOperation
{
private:
	CSize m_size;
public:
	CSize size() const;
	int width() const;
	int height() const;

	void resize(CSize newSize);

private:
	CPoint m_targetOrigin;
public:
	CPoint targetOrigin() const;
	CRect targetRect() const;
	int targetX() const;
	int targetY() const;

	void moveTo(CPoint newTargetOrigin);

public:
	explicit PatchOperation(CVclipDoc& owner, LONG targetFrame = 0,
		int targetX = 0, int targetY = 0, int width = 0, int height = 0);
};

// Reference counting

class PFrameOperation : public counted_ptr<FrameOperation>
{
private:
	typedef counted_ptr<FrameOperation> Base;
public:
	explicit PFrameOperation(FrameOperation* raw = 0) : Base(raw) {}
	PFrameOperation(const PFrameOperation& other) : Base(other) {}

	void accept(IOperationVisitor& visitor) const;
};

class byTargetFrame : public std::binary_function<PFrameOperation, PFrameOperation, bool>
{
public:
	bool operator()(const PFrameOperation& lhs, const PFrameOperation& rhs) const;
	bool operator()(const PFrameOperation& lhs, LONG rhs) const;
	bool operator()(LONG lhs, const PFrameOperation& rhs) const;
};

std::ostream& operator<<(std::ostream& stream, const PFrameOperation& op);

// Visitor

class FramePatch;
class FieldPatch;
class FrameImage;
class FieldRebuild;
class MotionInterpolate;

class IOperationVisitor
{
public:
	virtual void operator()(FramePatch& patch) = 0;
	virtual void operator()(FieldPatch& patch) = 0;
	virtual void operator()(FrameImage& image) = 0;
	virtual void operator()(FieldRebuild& rebuild) = 0;
	virtual void operator()(MotionInterpolate& inter) = 0;
};

// Factory

class ICreateFrameOperation
{
public:
	virtual PFrameOperation operator()(CVclipDoc& owner) const = 0;
};

template <typename OperationT>
class CreateFrameOperation : public ICreateFrameOperation
{
private: // noncopyable
	CreateFrameOperation(const CreateFrameOperation& other);
	CreateFrameOperation& operator=(const CreateFrameOperation& other);

public:
	explicit CreateFrameOperation(const char* typeCode);
	PFrameOperation operator()(CVclipDoc& owner) const;
};

class FrameOperationFactory
{
private: // singleton
	FrameOperationFactory();
	FrameOperationFactory(const FrameOperationFactory& other);
	FrameOperationFactory& operator=(const FrameOperationFactory& other);
public:
	static FrameOperationFactory& instance();

private: // serialization
	typedef std::map<std::string, const ICreateFrameOperation*> Creators;
	Creators m_creators;
	PFrameOperation create(const std::string& typeCode, CVclipDoc& owner) const;
public:
	void insert(const std::string& typeCode, const ICreateFrameOperation& creator);
	PFrameOperation read(std::istream& stream, CVclipDoc& owner) const;
	void write(std::ostream& stream, PFrameOperation pfo) const;

public:
	PFrameOperation createFramePatch(CVclipDoc& owner,
		LONG sourceFrame, int sourceX, int sourceY, int sourceWidth, int sourceHeight,
		LONG targetFrame, int targetX, int targetY, HOverlay mode) const;
	PFrameOperation createFieldPatch(CVclipDoc& owner,
		LONG sourceFrame, LONG sourceOrder, int sourceX, int sourceY, int width, int height,
		LONG targetFrame, LONG targetOrder, int targetX, int targetY, HOverlay mode) const;
	PFrameOperation createFrameImage(CVclipDoc& owner,
		LONG targetFrame, const std::string& fileName) const;
	PFrameOperation createFieldRebuild(CVclipDoc& owner, 
		LONG targetFrame) const;
	PFrameOperation createMotionInterpolate(CVclipDoc& owner, 
		LONG targetFrame) const;
};

template <typename T>
T& the()
{
	return T::instance();
};

#include "FrameOperation.inline.h"

