#pragma once

class ref_counted
{
private:
	mutable int m_refCount;
protected:
	ref_counted(const ref_counted& other)
		: m_refCount(0)
	{}
	ref_counted& operator=(const ref_counted& other)
	{
		// intentionally skip copying m_refCount;
	}
	virtual ~ref_counted() = 0;
public:
	ref_counted()
		: m_refCount(0)
	{}

	void add_ref() const
	{
		++m_refCount;
	}

	void release() const
	{
		if (!--m_refCount) delete this;
	}
};

inline ref_counted::~ref_counted() {}

template <typename CountedT>
class counted_ptr
{
private:
	CountedT* m_p;
	typedef CountedT* (counted_ptr::* unspecified_bool)() const;
public:
	explicit counted_ptr(CountedT* raw = 0)
		: m_p(raw)
	{
		if (m_p) m_p->add_ref();
	}

	counted_ptr(const counted_ptr& other)
		: m_p(other.m_p)
	{
		if (m_p) m_p->add_ref();
	}

	void swap(counted_ptr& other)
	{
		std::swap(m_p, other.m_p);
	}

	counted_ptr& operator=(const counted_ptr& other)
	{
		counted_ptr(other).swap(*this);
		return *this;
	}

	~counted_ptr()
	{
		reset();
	}

	void reset()
	{
		if (m_p) m_p->release();
		m_p = 0;
	}

	CountedT* get() const
	{
		return m_p;
	}

	CountedT& operator*() const
	{
		return *m_p;
	}

	CountedT* operator->() const
	{
		return m_p;
	}

	operator unspecified_bool() const
	{
		return m_p ? &counted_ptr::get : 0;
	}

	bool operator!() const
	{
		return !m_p;
	}

	friend operator==(const counted_ptr& lhs, const counted_ptr& rhs)
	{
		return lhs.m_p == rhs.m_p;
	}

	friend operator!=(const counted_ptr& lhs, const counted_ptr& rhs)
	{
		return !(lhs == rhs);
	}
};

namespace std
{
	template <typename CountedT>
	void swap(counted_ptr<CountedT>& lhs, counted_ptr<CountedT>& rhs)
	{
		lhs.swap(rhs);
	}
}

