// vclipDoc.cpp : implementation of the CVclipDoc class
//

#include "stdafx.h"
#include "vclip.h"

#include "vclipDoc.h"

#include "exception.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include <ios>
#include <istream>
#include <ostream>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <functional>
#include <exception>
#include <stdexcept>
#include ".\vclipdoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// CVclipDoc

IMPLEMENT_DYNCREATE(CVclipDoc, CDocument)

BEGIN_MESSAGE_MAP(CVclipDoc, CDocument)
END_MESSAGE_MAP()


// CVclipDoc construction/destruction

CVclipDoc::CVclipDoc()
{
}

CVclipDoc::~CVclipDoc()
{
}

BOOL CVclipDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	m_source.clear();
	m_operations.clear();

	SetModifiedFlag(FALSE);
	UpdateAllViews(0, -2);

	return TRUE;
}

BOOL CVclipDoc::OnOpenDocument(LPCTSTR lpszPathName)
{
	DeleteContents();
	CString context;
	try
	{
		std::ifstream stream;
		stream.exceptions(std::ios_base::badbit);

		context = "opening file";
		stream.open(lpszPathName);

		size_t lineNumber = 1;

		context = "reading source video file name from";
		std::getline(stream, m_source);
		++lineNumber;

		do
		{
			context.Format("reading line %d from", lineNumber);

			std::string line;
			std::getline(stream, line);
			if (!stream || line.empty()) break;

			context.Format("reading line %d (\"%s\") from", lineNumber, line.c_str());
	
			std::istringstream iss(line);
			PFrameOperation op = the<FrameOperationFactory>().read(iss, *this); 

			m_operations.push_back(op);
			++lineNumber;
		} while (true);

		SetModifiedFlag(FALSE);
		UpdateAllViews(0, -2);

		return TRUE;
	}
	catch (const std::exception& e)
	{
		DeleteContents();
		CString error;
		error.Format("An error occured while %s %s.\n%s", context, lpszPathName, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
		return FALSE;
	}
	catch (...)
	{
		DeleteContents();
		CString error;
		error.Format("An unknown error occured while %s %s.", context, lpszPathName);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
		return FALSE;
	}
}

BOOL CVclipDoc::OnSaveDocument(LPCTSTR lpszPathName)
{
	CString context;
	try
	{
		std::ofstream stream;
		stream.exceptions(std::ios_base::badbit | std::ios_base::failbit);

		context = "opening file";
		stream.open(lpszPathName);

		context = "writing video file name to";
		stream << m_source << "\n";

		context = "writing operations to";
		std::copy(m_operations.begin(), m_operations.end(),
			std::ostream_iterator<PFrameOperation>(stream, "\n"));

		SetModifiedFlag(FALSE);
		return TRUE;
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s %s.\n%s", context, lpszPathName, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
		return FALSE;
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s %s.", context, lpszPathName);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
		return FALSE;
	}
}

bool ends_with(const std::string& whole, const std::string& suffix)
{
	return std::equal(whole.end() - suffix.size(), whole.end(), suffix.begin());
}

int Compiler::roundDown(int n) const
{
	return m_evenCoords ? (n &~ 1) : n;
}

int Compiler::roundUp(int n) const
{
	return m_evenCoords ? (n + (n & 1)) : n;
}

Compiler::Compiler(std::ostream& stream, bool evenCoords, const CRect& frameSize)
	: m_stream(&stream), m_frame(-1), m_evenCoords(evenCoords), m_needPlus(false), m_needCR(false),
	m_frameSize(frameSize)
{
}

void Compiler::source(const std::string& source)
{
	if (ends_with(source, ".avs"))
	{
		std::ifstream ifs;
		ifs.exceptions(std::ios_base::badbit);

		ifs.open(source.c_str());

		std::string s;
		for (std::getline(ifs, s); ifs; std::getline(ifs, s))
		{
			*m_stream << s << "\n";
		}
		*m_stream << "\nsource = last\n\n";
	}
	else
	{
		*m_stream << "source = AviSource(\"" << source << "\").ConvertToYUY2(interlaced=true)\n\n";
	}

	{ // import library functions
		std::vector<char> basename(MAX_PATH);
		basename.resize(ASSERT_NE(GetModuleFileName(0, &basename[0], MAX_PATH), 0, "GetModuleFileName"));

		*m_stream <<
			"Import(\"" << std::string(basename.begin(), std::find(basename.rbegin(), basename.rend(), '.').base()) << "avsi\")\n";
	}

	*m_stream <<
		"padding = source.BlankClip(length = 1)\n"
		"padding + source + padding\n\n";
}

void Compiler::operator()(const PFrameOperation& op)
{
	op->accept(*this);
}

void Compiler::operator()(FramePatch& patch)
{
	if (patch.sourceRect() == m_frameSize)
	{
		if (patch.targetOrigin() == CPoint(0, 0) && patch.mode()->isOpaque())
		{
			*m_stream << "FrameReplace(" <<
				patch.sourceFrame() << ",  " <<
				patch.targetFrame() << ")\n";
		}
		else
		{
			*m_stream << "FramePatch_Full(" <<
				patch.sourceFrame() << ",  " <<
				patch.targetFrame() << ",  " << 
					roundDown(patch.targetX()) << ", " << roundDown(patch.targetY()) << ",  "
				"\"" << patch.mode() << "\")\n";
		}
	}
	else
	{
		*m_stream << "FramePatch(" << 
			patch.sourceFrame() << ",  " << 
				roundDown(patch.sourceX()) << ", " << roundDown(patch.sourceY()) << ", " << 
				roundUp(patch.width()) << ", " << roundUp(patch.height()) << ",  " <<
			patch.targetFrame() << ",  " << 
				roundDown(patch.targetX()) << ", " << roundDown(patch.targetY()) << ",  "
				"\"" << patch.mode() << "\")\n";
	}
}

void Compiler::operator()(FieldPatch& patch)
{
	if (patch.sourceRect() == m_frameSize)
	{
		if (patch.targetOrigin() == CPoint(0, 0) && patch.mode()->isOpaque())
		{
			*m_stream << "FieldReplace(" << 
				patch.sourceFrame() << ", " << patch.sourceOrder() << ",  " << 
				patch.targetFrame() << ", " << patch.targetOrder() << ")\n";
		}
		else
		{
			*m_stream << "FieldPatch_Full(" << 
				patch.sourceFrame() << ", " << patch.sourceOrder() << ",  " << 
				patch.targetFrame() << ", " << patch.targetOrder() << ",  " << 
					roundDown(patch.targetX()) << ", " << roundDown(patch.targetY() / 2) << ",  "
				"\"" << patch.mode() << "\")\n";
		}
	}
	else
	{
		*m_stream << "FieldPatch(" << 
			patch.sourceFrame() << ", " << patch.sourceOrder() << ",  " << 
				roundDown(patch.sourceX()) << ", " << roundDown(patch.sourceY() / 2) << ", " << 
				roundUp(patch.width()) << ", " << roundUp(patch.height() / 2) << ",  " <<
			patch.targetFrame() << ", " << patch.targetOrder() << ",  " << 
				roundDown(patch.targetX()) << ", " << roundDown(patch.targetY() / 2) << ",  "
				"\"" << patch.mode() << "\")\n";
	}
}

void Compiler::operator()(FrameImage& image)
{
	*m_stream << "FrameImage(" << image.targetFrame() << ", \"" << image.fileName() << "\")\n";
}

void Compiler::operator()(FieldRebuild& rebuild)
{
	*m_stream << "FieldRebuild(" << rebuild.targetFrame() << ")\n";
}

void Compiler::operator()(MotionInterpolate& inter)
{
	*m_stream << "MotionInterpolate(" << inter.targetFrame() << ")\n";
}

void Compiler::empty()
{
	*m_stream << "source\n";
}

void Compiler::end(int totalFrames)
{
	*m_stream << 
		"\n"
		"Trim(1, FrameCount - 2)\n";
}

void CVclipDoc::compile(LPCTSTR lpszPathName, bool evenCoords, int totalFrames, const CRect& frameSize)
{
	CString context;
	try
	{
		std::ofstream stream;
		stream.exceptions(std::ios_base::badbit | std::ios_base::failbit);

		context = "opening target file";
		stream.open(lpszPathName);

		context = "preparing for compilation of";
		Compiler compiler(stream, evenCoords, frameSize);

		context = "writing video source to";
		compiler.source(m_source);

		context = "compiling to";
		if (m_operations.empty())
		{
			compiler.empty();
			return;
		}
		std::for_each(m_operations.begin(), m_operations.end(),
			compiler).end(totalFrames);

		context = "closing compiled file";
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s %s.\n%s", context, lpszPathName, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s %s.", context, lpszPathName);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

const std::string& CVclipDoc::source() const
{
	return m_source;
}


void CVclipDoc::setSource(const std::string& source)
{
	SetModifiedFlag();
	m_source = source;
	UpdateAllViews(0, -1);
}

PFrameOperation CVclipDoc::add(const PFrameOperation& op)
{
	m_operations.insert(
		std::upper_bound(m_operations.begin(), m_operations.end(), 
			op, byTargetFrame()), 
		op);
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
	return op;
}

void CVclipDoc::erase(const PFrameOperation& op)
{
	m_operations.erase(
		std::remove(m_operations.begin(), m_operations.end(),
			op),
		m_operations.end());
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
}


CVclipDoc::const_range CVclipDoc::operations(LONG frame) const
{
	return std::equal_range(m_operations.begin(), m_operations.end(), 
		frame, byTargetFrame());
}

CVclipDoc::range CVclipDoc::operations(LONG frame)
{
	return std::equal_range(m_operations.begin(), m_operations.end(), 
		frame, byTargetFrame());
}

CVclipDoc::const_iterator CVclipDoc::begin() const
{
	return m_operations.begin();
}

CVclipDoc::const_iterator CVclipDoc::end() const
{
	return m_operations.end();
}

class usesFrame : public std::unary_function<PFrameOperation, bool>
{
private:
	LONG m_frame;
public:
	explicit usesFrame(LONG frame)
		: m_frame(frame)
	{
	}

	bool operator()(const PFrameOperation& op) const
	{
		return op->usesFrame(m_frame);
	}
};

bool CVclipDoc::needFrame(LONG frame) const
{
	return m_operations.end() != std::find_if(m_operations.begin(), m_operations.end(),
		usesFrame(frame) );
}

void CVclipDoc::sendToBack(PFrameOperation op)
{
	range eq = operations(op->targetFrame());
	iterator it = std::find(eq.first, eq.second, op);
	if (it == eq.second) throw std::logic_error("sendToBack: not our operation");
	*(std::copy_backward(eq.first, it, it + 1) - 1) = op;
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
}

bool CVclipDoc::canSendFurther(PFrameOperation op) const
{
	const_iterator low = std::lower_bound(m_operations.begin(), m_operations.end(), 
		op, byTargetFrame());
	return low != m_operations.end() && *low != op;
}

void CVclipDoc::sendFurther(PFrameOperation op)
{
	range eq = operations(op->targetFrame());
	iterator it = std::find(eq.first, eq.second, op);
	if (it == eq.first) throw std::logic_error("sendFurther: already at the bottom");
	if (it == eq.second) throw std::logic_error("sendFurther: not our operation");
	std::iter_swap(it, it - 1);
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
}

bool CVclipDoc::canBringNearer(PFrameOperation op) const
{
	const_iterator upper = std::upper_bound(m_operations.begin(), m_operations.end(), 
		op, byTargetFrame());
	return upper != m_operations.begin() && *(upper - 1) != op;
}

void CVclipDoc::bringNearer(PFrameOperation op)
{
	range eq = operations(op->targetFrame());
	iterator it = std::find(eq.first, eq.second, op);
	if (it == eq.second) throw std::logic_error("bringNearer: not our operation");
	if (it + 1 == eq.second) throw std::logic_error("bringNearer: already at the top");
	std::iter_swap(it, it + 1);
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
}

void CVclipDoc::bringToFront(PFrameOperation op)
{
	range eq = operations(op->targetFrame());
	iterator it = std::find(eq.first, eq.second, op);
	if (it == eq.second) throw std::logic_error("bringNearer: not our operation");
	*std::copy(it + 1, eq.second, it) = op;
	SetModifiedFlag();
	UpdateAllViews(0, op->targetFrame());
}


// CVclipDoc serialization

void CVclipDoc::Serialize(CArchive& ar)
{
}


// CVclipDoc diagnostics

#ifdef _DEBUG
void CVclipDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CVclipDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG
