// vclipView.cpp : implementation of the CVclipView class
//

#include "stdafx.h"
#include "vclip.h"

#include "vclipDoc.h"
#include "vclipView.h"
#include "GotoFrame.h"
#include "MediaTime.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include "TemporaryFile.h"

#include <windows.h>

#include <stdexcept>
#include <exception>
#include "exception.h"

#include <algorithm>
#include <functional>
#include <numeric>
#include <utility>

#include <cstdio>
#include <ios>
#include <sstream>
#include <fstream>

#include ".\vclipview.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

CRect normalizeRect(/*by value*/CRect rect)
{
	rect.NormalizeRect();
	return rect;
}

size_t intervalHitTest(LONG min, LONG max, LONG value)
{
	if (value <  min - 2) return 0;
	if (value <= min + 2) return 1;
	if (value <  max - 2) return 2;
	if (value <= max + 2) return 3;
	return 4;
}

LRESULT rectHitTest(CRect rect, CPoint point)
{
	static const LRESULT results[5][5] =
	{
		{ HTNOWHERE, HTNOWHERE,    HTNOWHERE, HTNOWHERE,     HTNOWHERE },
		{ HTNOWHERE, HTTOPLEFT,    HTTOP,     HTTOPRIGHT,    HTNOWHERE },
		{ HTNOWHERE, HTLEFT,       HTCLIENT,  HTRIGHT,       HTNOWHERE },
		{ HTNOWHERE, HTBOTTOMLEFT, HTBOTTOM,  HTBOTTOMRIGHT, HTNOWHERE },
		{ HTNOWHERE, HTNOWHERE,    HTNOWHERE, HTNOWHERE,     HTNOWHERE }
	};

	return results
		[intervalHitTest(rect.top,  rect.bottom, point.y)]
		[intervalHitTest(rect.left, rect.right,  point.x)];
}

//:View modes

CRect ViewMode::viewFromFrame(const CRect& rect) const
{
	return CRect(viewFromFrame(rect.TopLeft()), viewFromFrame(rect.BottomRight()));
}

CRect ViewMode::frameFromView(const CRect& rect) const
{
	return CRect(frameFromView(rect.TopLeft()), frameFromView(rect.BottomRight()));
}

class FrameMode : public ViewMode
{
public:
	CPoint viewFromFrame(const CPoint& point) const
	{
		return point;
	}
	CPoint frameFromView(const CPoint& point) const
	{
		return point;
	}

	LONG next(LONG& frame, LONG& /*order*/) const
	{
		return ++frame;
	}
	LONG previous(LONG& frame, LONG& /*order*/) const
	{
		return --frame;
	}
	bool hasNext(LONG frame, LONG /*order*/, LONG frames) const
	{
		return frame + 1 < frames;
	}
	bool hasPrevious(LONG frame, LONG /*order*/) const
	{
		return frame > 0;
	}
	CString formatPosition(LONG frame, LONG /*order*/, LONG frames) const
	{
		CString result;
		result.Format("Frame %d of %d", frame, frames);
		return result;
	}

	bool visible(LONG /*currentOrder*/, LONG /*targetOrder*/) const
	{
		return true;
	}

	void draw(CDC* pDC, const DrawDib& drawDib, const PFrame& frame, LONG /*order*/) const
	{
		drawDib.draw(
			pDC,   frame.rect(),
			frame, frame.rect(),
			0);
	}

	PFrameOperation paste(CVclipDoc& document, const Clipboard& clipboard, LONG frame, LONG order, HOverlay mode) const
	{
		return document.add(
			the<FrameOperationFactory>().createFramePatch(document,
				clipboard.frame, clipboard.selection.left, clipboard.selection.top,
				clipboard.selection.right - clipboard.selection.left, clipboard.selection.bottom - clipboard.selection.top,
				frame, clipboard.selection.left, clipboard.selection.top, mode));
	}
} frameMode;

class FieldMode : public ViewMode
{
public:
	CPoint viewFromFrame(const CPoint& point) const
	{
		return CPoint(point.x, (point.y - (point.y % 2 ? 1 : 0)) / 2);
	}
	CPoint frameFromView(const CPoint& point) const
	{
		return CPoint(point.x, point.y * 2);
	}
	using ViewMode::viewFromFrame;
	using ViewMode::frameFromView;

	LONG next(LONG& frame, LONG& order) const
	{
		switch (order)
		{
		case 0:
			order = 1;
			return frame;
		case 1:
			order = 0;
			return ++frame;
		}
		return frame;
	}
	LONG previous(LONG& frame, LONG& order) const
	{
		switch (order)
		{
		case 1:
			order = 0;
			return frame;
		case 0:
			order = 1;
			return --frame;
		}
		return frame;
	}
	bool hasNext(LONG frame, LONG order, LONG frames) const
	{
		return frame + order < frames;
	}
	bool hasPrevious(LONG frame, LONG order) const
	{
		return frame + order > 0;
	}
	CString formatPosition(LONG frame, LONG order, LONG frames) const
	{
		CString result;
		result.Format("Frame %d%c of %d", frame, order["bt"], frames);
		return result;
	}
	
	bool visible(LONG currentOrder, LONG targetOrder) const
	{
		return currentOrder == targetOrder;
	}

	void draw(CDC* pDC, const DrawDib& drawDib, const PFrame& frame, LONG order) const
	{
		PFrame fields(frame, SeparateFields());
		CRect fieldRect(0, 0, frame.width(), fields.height());
		drawDib.draw(
			pDC,    fieldRect,
			fields, fieldRect + fields.fieldOrigin(order),
			0);
	}

	PFrameOperation paste(CVclipDoc& document, const Clipboard& clipboard, LONG frame, LONG order, HOverlay mode) const
	{
		return document.add(
			the<FrameOperationFactory>().createFieldPatch(document,
			clipboard.frame, clipboard.order, clipboard.selection.left, clipboard.selection.top,
				clipboard.selection.right - clipboard.selection.left, clipboard.selection.bottom - clipboard.selection.top,
				frame, order, clipboard.selection.left, clipboard.selection.top, mode));
	}
} fieldMode;

void CVclipView::OnUpdateViewFramemode(CCmdUI *pCmdUI)
{
	pCmdUI->SetRadio(m_viewMode == &frameMode);
}
void CVclipView::OnViewFramemode()
{
	m_viewMode = &frameMode;
	Invalidate();
}

void CVclipView::OnUpdateViewFieldmode(CCmdUI *pCmdUI)
{
	pCmdUI->SetRadio(m_viewMode == &fieldMode);
}
void CVclipView::OnViewFieldmode()
{
	m_viewMode = &fieldMode;
	Invalidate();
}

//#View modes

//:Drag states

class NullDrag : public IDragState
{
public:
	void clipCursor(/*by value*/ CRect rect) const {}
	bool mouseMove(const CPoint& point) const { return false; }
	void mouseUp(const CPoint& point) const {}
	bool canCancel() const { return false; }
	void dragCancel() const {}
};

class PatchMove : public IDragState
{
private:
	PatchOperation* m_patch;
	CPoint m_offset;
	CPoint m_backup;
public:
	PatchMove(PatchOperation& patch, const CPoint& point)
		: m_patch(&patch), m_offset(point - patch.targetOrigin()), m_backup(patch.targetOrigin())
	{}

	void clipCursor(/*by value*/ CRect rect) const
	{
		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		m_patch->moveTo(point - m_offset);
		return false;
	}

	void mouseUp(const CPoint& point) const
	{
		m_patch->moveTo(point - m_offset);
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true;
	}
	void dragCancel() const
	{
		m_patch->moveTo(m_backup);
		ClipCursor(0);
	}
};

class SelectionMove : public IDragState
{
private:
	CRect* m_pRect;
	CRect m_offset;
	CRect m_backup;
public:
	SelectionMove(CPoint point, CRect& rect)
		: m_pRect(&rect), m_offset(rect - point), m_backup(rect)
	{}

	void clipCursor(/*by value*/CRect rect) const
	{
		rect.DeflateRect(-m_offset.left, -m_offset.top, m_offset.right, m_offset.bottom);
		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		*m_pRect = m_offset + point;
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		*m_pRect = m_offset + point;
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		*m_pRect = m_backup;
		ClipCursor(0);
	}
};

template <LONG CPoint::* member, LONG CRect::* low, LONG CRect::* high>
class SelectionSideResize : public IDragState
{
private:
	LONG* m_pTarget;
	LONG m_offset;
	LONG m_backup;
public:
	SelectionSideResize(CPoint point, LONG& target)
		: m_pTarget(&target), m_offset(target - point.*member), m_backup(target)
	{}

	void clipCursor(/*by value*/ CRect rect) const
	{
		if (m_offset < 0) rect.*low -= m_offset;
		if (m_offset > 0) rect.*high -= m_offset;
		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		*m_pTarget = m_offset + point.*member;
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		*m_pTarget = m_offset + point.*member;
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		*m_pTarget = m_backup;
		ClipCursor(0);
	}
};
typedef SelectionSideResize<&CPoint::x, &CRect::left, &CRect::right> SelectionHorizontalResize;
typedef SelectionSideResize<&CPoint::y, &CRect::top, &CRect::bottom> SelectionVerticalResize;

class SelectionCornerResize : public IDragState
{
private:
	LONG* m_px;
	LONG* m_py;
	CPoint m_offset;
	CPoint m_backup;
public:
	SelectionCornerResize(CPoint point, LONG& x, LONG& y)
		: m_px(&x), m_py(&y), m_offset(CPoint(x, y) - point), m_backup(x, y)
	{}

	void clipCursor(/*by value*/ CRect rect) const
	{
		if (m_offset.x < 0) rect.left -= m_offset.x;
		if (m_offset.y < 0) rect.top -= m_offset.y;
		if (m_offset.x > 0) rect.right -= m_offset.x;
		if (m_offset.y > 0) rect.bottom -= m_offset.y;
		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		*m_px = m_offset.x + point.x;
		*m_py = m_offset.y + point.y;
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		*m_px = m_offset.x + point.x;
		*m_py = m_offset.y + point.y;
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		*m_px = m_backup.x;
		*m_py = m_backup.y;
		ClipCursor(0);
	}
};

void CVclipView::DragCancel()
{
	m_dragState->dragCancel();
	m_dragState = std::auto_ptr<IDragState>(new NullDrag);
}

void CVclipView::EndDrag(CPoint point)
{
	m_dragState->mouseUp(point);
	m_dragState = std::auto_ptr<IDragState>(new NullDrag);
	m_selection.NormalizeRect();
	Invalidate();
}

//#Drag states

//:Postprocessors

template <typename T>
class bitwise_not : public std::unary_function<T, T>
{
public:
	T operator()(T value) const
	{
		return ~value;
	}
};

class Identity : public IPostProcessor
{
public:
	void operator()(PFrame /*frame*/) const
	{
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return worker;
	}
} const identity;

class Invert : public IPostProcessor
{
public:
	void operator()(PFrame frame) const
	{
		BYTE* data = reinterpret_cast<BYTE*>(frame.data());
		std::transform(data, data + frame.header()->biSizeImage,
			data,
			bitwise_not<BYTE>());
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return identity;
	}
} const invert;

class Deinterlace : public IPostProcessor
{
private:
	RGBTRIPLE blend(const RGBTRIPLE& a, const RGBTRIPLE& b, const RGBTRIPLE& c) const
	{
		RGBTRIPLE result;
		result.rgbtBlue = (a.rgbtBlue + 2 * b.rgbtBlue + c.rgbtBlue) / 4;
		result.rgbtGreen = (a.rgbtGreen + 2 * b.rgbtGreen + c.rgbtGreen) / 4;
		result.rgbtRed = (a.rgbtRed + 2 * b.rgbtRed + c.rgbtRed) / 4;
		return result;
	}
public:
	void operator()(PFrame frame) const
	{
		PFrame copy = frame.clone();
		for (LONG x = 0; x < frame.width(); ++x)
		{
			frame.at(x, 0) = blend(copy.at(x, 1), copy.at(x, 0), copy.at(x, 1));
		}
		for (LONG y = 1; y + 1 < copy.height(); ++y)
		{
			for (LONG x = 0; x < copy.width(); ++x)
			{
				frame.at(x, y) = blend(copy.at(x, y - 1), copy.at(x, y), copy.at(x, y + 1));
			}
		}
		for (LONG x = 0; x < copy.width(); ++x)
		{
			frame.at(x, copy.height() - 1) = blend(copy.at(x, copy.height() - 2), copy.at(x, copy.height() - 1), copy.at(x, copy.height() - 2));
		}
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return identity;
	}
} const deinterlace;

void CVclipView::OnUpdateViewInvert(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_invert == &invert);
}
void CVclipView::OnViewInvert()
{
	m_invert = m_invert->select(&identity, &invert);
	Invalidate();
}

void CVclipView::OnUpdateViewDeinterlace(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_deinterlace == &deinterlace);
}
void CVclipView::OnViewDeinterlace()
{
	m_deinterlace = m_deinterlace->select(&identity, &deinterlace);
	Invalidate();
}

//#Postprocessors

//:Drawing

class setTextColor
{
private:
	HDC m_hDC;
	COLORREF m_oldColor;
public:
	setTextColor(HDC hDC, COLORREF color)
		: m_hDC(hDC), m_oldColor(ASSERT_NE(SetTextColor(hDC, color), CLR_INVALID, "SetTextColor"))
	{
	}

	~setTextColor()
	{
		SetTextColor(m_hDC, m_oldColor);
	}
};

class setBkMode
{
private:
	HDC m_hDC;
	int m_oldMode;
public:
	setBkMode(HDC hDC, int mode)
		: m_hDC(hDC), m_oldMode(ASSERT_NE(SetBkMode(hDC, mode), 0, "SetBkMode"))
	{
	}

	~setBkMode()
	{
		SetBkMode(m_hDC, m_oldMode);
	}
};

void drawOutlinedText(CDC* pDC, COLORREF textColor, COLORREF outlineColor,
											const CString& string, CRect rect, UINT format)
{
	setBkMode sbm(*pDC, TRANSPARENT);

	setTextColor stcOutline(*pDC, outlineColor);

	ASSERT_NE(pDC->DrawText(string, rect + CPoint(-1, -1), format), 0, "DrawText(-1, -1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint( 1, -1), format), 0, "DrawText( 1, -1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint(-1,  1), format), 0, "DrawText(-1,  1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint( 1,  1), format), 0, "DrawText( 1,  1)");

	setTextColor stcText(*pDC, textColor);

	ASSERT_NE(pDC->DrawText(string, rect, format), 0, "DrawText");
}

class CanDrawDirectly : public std::binary_function<bool, PFrameOperation, bool>,
	public IOperationVisitor
{
private:
	bool m_canDraw;
public:
	bool operator()(bool accum, const PFrameOperation& op)
	{
		if (!accum) return false;
		op.accept(*this);
		return m_canDraw;
	}
	
	void operator()(FramePatch& /*patch*/)        { m_canDraw = true;  }
	void operator()(FieldPatch& /*patch*/)        { m_canDraw = true;  }
	void operator()(FrameImage& /*image*/)        { m_canDraw = true;  }
	void operator()(FieldRebuild& /*patch*/)      { m_canDraw = true;  }
	void operator()(MotionInterpolate& /*inter*/) { m_canDraw = false; }
};

class DrawOperationInFrame : public std::unary_function<PFrameOperation, void>,
	public IOperationVisitor
{
private:
	CDC* m_pDC;
	PFrame* m_pFrame;
	const AviFile* m_pAviFile;
	const ImageCache* m_pImageCache;
public:
	DrawOperationInFrame(CDC* pDC, PFrame& frame, const AviFile& aviFile, const ImageCache& imageCache)
		: m_pDC(pDC), m_pFrame(&frame), m_pAviFile(&aviFile), m_pImageCache(&imageCache)
	{
	}

	void operator()(const PFrameOperation& op)
	{
		op.accept(*this);
	}

	void operator()(FramePatch& patch)
	{
		m_pFrame->copyFrom(
			(*m_pAviFile)[patch.sourceFrame()], 
			patch.sourceRect(), 
			patch.targetRect(),
			*patch.mode());
	}

	void operator()(FieldPatch& patch)
	{
		PFrame(*m_pFrame, SeparateFields()).copyFrom(
			PFrame((*m_pAviFile)[patch.sourceFrame()], SeparateFields()), 
			fieldMode.viewFromFrame(patch.sourceRect()) + CPoint(patch.sourceOrder() * m_pFrame->width(), 0),
			fieldMode.viewFromFrame(patch.targetRect()) + CPoint(patch.targetOrder() * m_pFrame->width(), 0),
			*patch.mode());
	}

	void operator()(FrameImage& image)
	{
		PBitmap bitmap = (*m_pImageCache)[image.fileName()];
		m_pFrame->copyFrom(bitmap);
	}

	void DrawOperationInFrame::operator()(FieldRebuild& rebuild)
	{
		PFrame fields(*m_pFrame, SeparateFields());
		PFrame previous((*m_pAviFile)[rebuild.targetFrame() - 1], SeparateFields());
		PFrame next((*m_pAviFile)[rebuild.targetFrame() + 1], SeparateFields());

		CRect fieldRect(0, 0, m_pFrame->width(), fields.height());

		fields.copyFrom(previous, fieldRect + previous.fieldOrigin(0), fieldRect + fields.fieldOrigin(0));
		fields.copyFrom(next,     fieldRect + next    .fieldOrigin(1), fieldRect + fields.fieldOrigin(1));
	}

	void operator()(MotionInterpolate& rebuild)
	{
		CDC dc;
		ASSERT_NE(dc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");

		CBitmap bitmap;
		ASSERT_NE(bitmap.CreateCompatibleBitmap(m_pDC, m_pFrame->width(), m_pFrame->height()), FALSE, "CreateCompatibleBitmap");
		ASSERT_NE(SetDIBits(*m_pDC, bitmap, 0, m_pFrame->height(), m_pFrame->data(), 
			reinterpret_cast<LPBITMAPINFO>(m_pFrame->header()), DIB_RGB_COLORS), 0, "SetDIBits");

		CFont font;
		ASSERT_NE(font.CreateFont(
			0, 0, // width, height
			0, 0, // escapement, orientation
			FW_BOLD, FALSE, FALSE, 0, // weight, italic, underline, strikeout
			ANSI_CHARSET,
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
			DEFAULT_PITCH,
			"MS Shell Dlg"), FALSE, "CreateFont");

		ASSERT_NE(dc.SelectObject(bitmap), HGDIOBJ(), "SelectObject(bitmap)");
		ASSERT_NE(dc.SelectObject(font), HGDIOBJ(), "SelectObject(font)");

		CRect textRect = m_pFrame->rect();
		textRect.DeflateRect(10, 10);

		drawOutlinedText(&dc, RGB(255, 255, 0), RGB(0, 0, 0),
			"This frame will be motion-interpolated.\n"
			"Vclip doesnt know how to display motion-interpolated frames, "
			"but can call upon the powers of Avisynth to render them. "
			"Select View | Accurate Rendering if you wish to see the exact "
			"Avisynth output. But please be aware that it may be slow.",
			textRect, DT_LEFT | DT_TOP | DT_NOPREFIX | DT_WORDBREAK);

		ASSERT_NE(GetDIBits(*m_pDC, bitmap, 0, m_pFrame->height(), m_pFrame->data(), 
			reinterpret_cast<LPBITMAPINFO>(m_pFrame->header()), DIB_RGB_COLORS), 0, "GetDIBits");
	}
};

PFrame CVclipView::PrepareFrame(CDC* pDC) const
{
	PFrame frame = m_aviFile.getUncached(m_pos);
	if (!frame) return frame;

	CVclipDoc::range range = GetDocument()->operations(m_pos);

	if (m_accurateRendering && 
		!std::accumulate(range.first, range.second, true, CanDrawDirectly()))
	{
		TemporaryFilename avsName(".avs");

		std::ofstream avs(avsName.c_str());
		Compiler compiler(avs, m_evenCoords, m_aviFile.frameSize());
		compiler.source(GetDocument()->source());
		std::for_each(range.first, range.second, compiler);
		compiler.end(m_pos);
		avs.close();

		AviFile avsAviFile;
		avsAviFile.open(avsName.c_str());
		frame = avsAviFile.getUncached(m_pos).clone();
	}
	else
	{
		std::for_each(range.first, range.second, 
			DrawOperationInFrame(pDC, frame, m_aviFile, m_imageCache));
	}

	(*m_invert)(frame);
	(*m_deinterlace)(frame);

	return frame;
}

void CVclipView::OnUpdateViewAccuraterendering(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
	pCmdUI->SetCheck(m_accurateRendering);
}
void CVclipView::OnViewAccuraterendering()
{
	m_accurateRendering = !m_accurateRendering;
	Invalidate();
}

void CVclipView::OnDraw(CDC* pDC)
{
	CVclipDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	if (!pDoc)
		return;

	CString context = "drawing";
	try
	{
		if (PFrame frame = PrepareFrame(pDC))
		{
			m_viewMode->draw(pDC, m_drawDib, frame, m_order);
		}

		context.Format("drawing active operation");

		m_manipulator->draw(pDC, m_viewMode);
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//#Drawing

//:Manipulators

class NullManipulator : public IManipulator
{
	void draw(CDC* pDC, const ViewMode* mode) const {}
	bool wantClick(const CPoint& point) const { return false; }
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		return std::auto_ptr<IDragState>(new NullDrag); 
	}
	bool canErase() const { return false; }
	void erase(CVclipDoc& document) const {}
	void accept(IOperationVisitor& visitor) const {}
	bool setCursor(const CPoint& point) const { return false; }
	bool isOperation() const { return false; }
	PFrameOperation operation() const { return PFrameOperation(); }
	bool canMove(int dx, int dy) const { return false; }
	void moveBy(int dx, int dy) const {}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class SelectionManipulator : public IManipulator
{
private:
	const ViewMode* m_viewMode;
	CRect* m_pSelection;
	CRect m_frameRect;
public:
	explicit SelectionManipulator(const ViewMode* viewMode, CRect& selection, const CRect& frameRect)
		: m_viewMode(viewMode), m_pSelection(&selection), m_frameRect(frameRect)
	{}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		CRect frameRect = normalizeRect(*m_pSelection);
		if (frameRect.IsRectEmpty()) return;
		pDC->DrawFocusRect(mode->viewFromFrame(frameRect));
	}

	bool wantClick(const CPoint& point) const
	{
		return !!m_pSelection->PtInRect(point);
	}

	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		CRect viewRect = m_viewMode->viewFromFrame(*m_pSelection);
		CPoint viewPoint = m_viewMode->viewFromFrame(point);
		switch (rectHitTest(viewRect, viewPoint))
		{
		#define AND ,
		#define YK_HITTEST(code, handlerClass, params) \
		case code: return std::auto_ptr<IDragState>(new Selection##handlerClass##Resize(point, params));

			YK_HITTEST(HTTOPLEFT    , Corner    , m_pSelection->left  AND m_pSelection->top   );
			YK_HITTEST(HTTOP        , Vertical  ,                         m_pSelection->top   );
			YK_HITTEST(HTTOPRIGHT   , Corner    , m_pSelection->right AND m_pSelection->top   );
			YK_HITTEST(HTLEFT       , Horizontal, m_pSelection->left                          );
			YK_HITTEST(HTRIGHT      , Horizontal, m_pSelection->right                         );
			YK_HITTEST(HTBOTTOMLEFT , Corner    , m_pSelection->left  AND m_pSelection->bottom);
			YK_HITTEST(HTBOTTOM     , Vertical  ,                         m_pSelection->bottom);
			YK_HITTEST(HTBOTTOMRIGHT, Corner    , m_pSelection->right AND m_pSelection->bottom);

		#undef YK_HITTEST
		#undef AND
		case HTCLIENT:
			return std::auto_ptr<IDragState>(new SelectionMove(point, *m_pSelection));
		default:
			*m_pSelection = CRect(point, point);
			return std::auto_ptr<IDragState>(new SelectionCornerResize(point,
				m_pSelection->right, m_pSelection->bottom));
		}
	}

	bool canErase() const { return false; }
	void erase(CVclipDoc& /*document*/) const {}
	void accept(IOperationVisitor& visitor) const {}

	bool setCursor(const CPoint& point) const
	{
		CRect viewRect = m_viewMode->viewFromFrame(*m_pSelection);
		CPoint viewPoint = m_viewMode->viewFromFrame(point);
		switch (rectHitTest(viewRect, viewPoint))
		{
		case HTTOPLEFT:     case HTBOTTOMRIGHT:  SetCursor(LoadCursor(0, IDC_SIZENWSE)); return true;
		case HTTOP:         case HTBOTTOM:       SetCursor(LoadCursor(0, IDC_SIZENS));   return true;
		case HTTOPRIGHT:    case HTBOTTOMLEFT:   SetCursor(LoadCursor(0, IDC_SIZENESW)); return true;
		case HTLEFT:        case HTRIGHT:        SetCursor(LoadCursor(0, IDC_SIZEWE));   return true;
		case HTCLIENT:                           SetCursor(LoadCursor(0, IDC_SIZEALL));  return true;
		}
		return false;
	}

	bool isOperation() const { return false; }
	PFrameOperation operation() const { return PFrameOperation(); }

	bool canMove(int dx, int dy) const
	{
		return m_frameRect.PtInRect(m_pSelection->TopLeft() + CPoint(dx, dy)) &&
			m_frameRect.PtInRect(m_pSelection->BottomRight() + CPoint(dx, dy));
	}
	void moveBy(int dx, int dy) const
	{
		m_pSelection->OffsetRect(dx, dy);
	}
	bool canSize(int dx, int dy) const
	{
		return !!m_frameRect.PtInRect(m_pSelection->BottomRight() + CPoint(dx, dy));
	}
	void sizeBy(int dx, int dy) const
	{
		m_pSelection->right += dx;
		m_pSelection->bottom += dy;
		m_pSelection->NormalizeRect();
	}
};

class FullFrameManipulator : public IManipulator
{
private:
	PFrameOperation m_operation;
	CRect m_frameRect;
public:
	explicit FullFrameManipulator(PFrameOperation operation, const CRect& frameRect)
		: m_operation(operation), m_frameRect(frameRect)
	{}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		pDC->DrawFocusRect(mode->viewFromFrame(m_frameRect));
	}
	bool wantClick(const CPoint& point) const
	{
		return true;
	}
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		return std::auto_ptr<IDragState>(new NullDrag);
	}
	bool canErase() const { return true; }
	void erase(CVclipDoc& document) const { document.erase(m_operation); }
	void accept(IOperationVisitor& visitor) const { m_operation->accept(visitor); }
	bool setCursor(const CPoint& point) const
	{
		SetCursor(LoadCursor(0, IDC_ARROW));
		return true;
	}
	bool isOperation() const { return true; }
	PFrameOperation operation() const { return m_operation; }
	bool canMove(int dx, int dy) const { return false; }
	void moveBy(int dx, int dy) const {}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class PatchManipulator : public IManipulator
{
private:
	PFrameOperation m_operation;
	PatchOperation* m_patch;
	CRect m_frameRect;
public:
	explicit PatchManipulator(PatchOperation& patch, const CRect& frameRect)
		: m_operation(&patch), m_patch(&patch), m_frameRect(frameRect)
	{}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		pDC->DrawFocusRect(mode->viewFromFrame(normalizeRect(m_patch->targetRect())));
	}
	bool wantClick(const CPoint& point) const
	{
		return !!m_patch->targetRect().PtInRect(point);
	}
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		return std::auto_ptr<IDragState>(new PatchMove(*m_patch, point));
	}
	bool canErase() const { return true; }
	void erase(CVclipDoc& document) const { document.erase(PFrameOperation(m_patch)); }
	void accept(IOperationVisitor& visitor) const { m_patch->accept(visitor); }
	bool setCursor(const CPoint& point) const
	{
		if (m_patch->targetRect().PtInRect(point))
		{
			SetCursor(LoadCursor(0, IDC_SIZEALL));
			return true;
		}
		return false;
	}
	bool isOperation() const { return true; }
	PFrameOperation operation() const { return PFrameOperation(m_patch); }
	bool canMove(int dx, int dy) const
	{
		// condition: at least one point must be in the frame

		return !(m_frameRect & (m_patch->targetRect() + CPoint(dx, dy))).IsRectEmpty();
	}
	void moveBy(int dx, int dy) const
	{
		m_patch->moveTo(m_patch->targetOrigin() + CPoint(dx, dy));
	}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class MakeManipulator : public IOperationVisitor
{
private:
	std::auto_ptr<IManipulator>* m_manipulator;
	CRect m_frameRect;
public:
	explicit MakeManipulator(std::auto_ptr<IManipulator>& pManipulator, const CRect& frameRect)
		: m_manipulator(&pManipulator), m_frameRect(frameRect)
	{}

	void operator()(FramePatch& patch)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new PatchManipulator(patch, m_frameRect));
	}
	void operator()(FieldPatch& patch)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new PatchManipulator(patch, m_frameRect));
	}
	void operator()(FrameImage& image)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&image), m_frameRect));
	}
	void operator()(FieldRebuild& rebuild)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&rebuild), m_frameRect));
	}
	void operator()(MotionInterpolate& inter)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&inter), m_frameRect));
	}
};

class GetOverlayMode : public IOperationVisitor
{
private:
	HOverlay* m_pMode;
public:
	explicit GetOverlayMode(HOverlay& mode)
		: m_pMode(&mode)
	{}

	void operator()(FramePatch& patch)
	{
		*m_pMode = patch.mode();
	}
	void operator()(FieldPatch& patch)
	{
		*m_pMode = patch.mode();
	}
	void operator()(FrameImage& image) {}
	void operator()(FieldRebuild& rebuild) {}
	void operator()(MotionInterpolate& inter) {}
};

void CVclipView::UnselectOperation()
{
	m_manipulator = std::auto_ptr<IManipulator>
		(new SelectionManipulator(m_viewMode, m_selection, m_aviFile.frameSize()));
	m_historyList.SetCurSel(-1);
	Invalidate();
}

void CVclipView::SelectOperation(PFrameOperation op)
{
	if (!op)
	{
		UnselectOperation();
	}
	else
	{
		op.accept(MakeManipulator(m_manipulator, m_aviFile.frameSize()));
		op.accept(GetOverlayMode(m_overlayMode));

		int i;
		for (i = 0; i < m_historyList.GetCount() && m_historyList.GetItemDataPtr(i) != op.get(); ++i);
		m_historyList.SetCurSel(i);

		m_selection.SetRectEmpty();

		Invalidate();
	}
}

class HitTest : public std::unary_function<PFrameOperation, void>, public IOperationVisitor
{
private:
	const ViewMode* m_viewMode;
	LONG m_order;
	CPoint m_framePoint;
	PFrameOperation* m_operation;
public:
	explicit HitTest(const ViewMode* viewMode, LONG order, 
		const CPoint& framePoint, PFrameOperation& operation)
	: m_viewMode(viewMode), m_order(order), 
		m_framePoint(framePoint), m_operation(&operation)
	{
	}

	void operator()(FramePatch& patch)
	{
		if (patch.targetRect().PtInRect(m_framePoint))
		{
			*m_operation = PFrameOperation(&patch);
		}
	}

	void operator()(FieldPatch& patch)
	{
		if (patch.targetRect().PtInRect(m_framePoint) && 
			m_viewMode->visible(patch.targetOrder(), m_order))
		{
			*m_operation = PFrameOperation(&patch);
		}
	}

	void operator()(FrameImage& image)
	{
		*m_operation = PFrameOperation(&image);
	}

	void operator()(FieldRebuild& rebuild)
	{
		*m_operation = PFrameOperation(&rebuild);
	}

	void operator()(MotionInterpolate& inter)
	{
		*m_operation = PFrameOperation(&inter);
	}

	void operator()(const PFrameOperation& op)
	{
		op.accept(*this);
	}
};

void CVclipView::OnLButtonDown(UINT nFlags, CPoint viewPoint)
{
	CView::OnLButtonDown(nFlags, viewPoint);

	if (!m_aviFile.size()) return;

	CRect clip;
	GetClientRect(&clip);
	clip.bottom -= 20;
	clip &= m_viewMode->viewFromFrame(m_aviFile.frameSize());
	++clip.bottom; ++clip.right;

	if (!clip.PtInRect(viewPoint)) return;

	CPoint framePoint = m_viewMode->frameFromView(viewPoint);

	CVclipDoc::range range = GetDocument()->operations(m_pos);

	if (MK_CONTROL & nFlags)
	{
		UnselectOperation();
	}
	else if (m_manipulator->wantClick(framePoint))
	{
		// the current manipulator stays
	}
	else
	{
		UnselectOperation();
		PFrameOperation clickedOperation;
		std::for_each(range.first, range.second,
			HitTest(m_viewMode, m_order, framePoint, clickedOperation));
		SelectOperation(clickedOperation);
	}

	ClientToScreen(&clip);
	m_dragState = m_manipulator->mouseDown(framePoint);
	m_dragState->clipCursor(clip);

	Invalidate();
}


void CVclipView::OnMouseMove(UINT nFlags, CPoint viewPoint)
{
	CView::OnMouseMove(nFlags, viewPoint);

	if (m_dragState->mouseMove(m_viewMode->frameFromView(viewPoint)))
	{
		Invalidate();
	}
}

void CVclipView::OnLButtonUp(UINT nFlags, CPoint viewPoint)
{
	CView::OnLButtonUp(nFlags, viewPoint);

	EndDrag(m_viewMode->frameFromView(viewPoint));
}

void CVclipView::OnDragcancel()
{
	if (m_dragState->canCancel())
	{
		DragCancel();
	}
	else
	{
		UnselectOperation();
		m_selection.SetRectEmpty();
	}
}

void CVclipView::OnUpdateEditMoveleft(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(-1, 0));
}
void CVclipView::OnEditMoveleft()
{
	m_manipulator->moveBy(-1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditMoveright(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(1, 0));
}
void CVclipView::OnEditMoveright()
{
	m_manipulator->moveBy(1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditMoveup(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(0, -1));
}
void CVclipView::OnEditMoveup()
{
	m_manipulator->moveBy(0, -1);
	Invalidate();
}

void CVclipView::OnUpdateEditMovedown(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(0, 1));
}
void CVclipView::OnEditMovedown()
{
	m_manipulator->moveBy(0, 1);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeleft(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(-1, 0));
}
void CVclipView::OnEditSizeleft()
{
	m_manipulator->sizeBy(-1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeright(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(1, 0));
}
void CVclipView::OnEditSizeright()
{
	m_manipulator->sizeBy(1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeup(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(0, -1));
}
void CVclipView::OnEditSizeup()
{
	m_manipulator->sizeBy(0, -1);
	Invalidate();
}

void CVclipView::OnUpdateEditSizedown(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(0, 1));
}
void CVclipView::OnEditSizedown()
{
	m_manipulator->sizeBy(0, 1);
	Invalidate();
}

void CVclipView::OnUpdateEditSelectnext(CCmdUI *pCmdUI)
{
	CVclipDoc::range range = GetDocument()->operations(m_pos);
	pCmdUI->Enable(!range.empty() &&
		*(range.second - 1) != m_manipulator->operation());
}
void CVclipView::OnEditSelectnext()
{
	CVclipDoc::range range = GetDocument()->operations(m_pos);
	SelectOperation(m_manipulator->isOperation() ? 
		*(std::find(range.first, range.second, m_manipulator->operation()) + 1) :
		*range.first);
}

void CVclipView::OnUpdateEditSelectprevious(CCmdUI *pCmdUI)
{
	CVclipDoc::range range = GetDocument()->operations(m_pos);
	pCmdUI->Enable(!range.empty() &&
		*(range.first) != m_manipulator->operation());
}
void CVclipView::OnEditSelectprevious()
{
	CVclipDoc::range range = GetDocument()->operations(m_pos);
	SelectOperation(m_manipulator->isOperation() ?
		*(std::find(range.first, range.second, m_manipulator->operation()) - 1) :
		*range.first);
}

//#Manipulators

//:Edit

void Clipboard::copy(const ViewMode* aMode, LONG aFrame, LONG anOrder, const CRect& aSelection)
{
	mode = aMode;
	frame = aFrame;
	order = anOrder;
	selection = aSelection;
}

PFrameOperation Clipboard::paste(CVclipDoc& document, const ViewMode* aMode, 
																 LONG aFrame, LONG anOrder, HOverlay overlayMode) const
{
	return mode->paste(document, *this, aFrame, anOrder, overlayMode);
}

bool Clipboard::empty() const
{
	return !!selection.IsRectEmpty();
}

void CVclipView::OnUpdateEditCopy(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(!m_selection.IsRectEmpty());
}
void CVclipView::OnEditCopy()
{
	m_clipboard.copy(m_viewMode, m_pos, m_order, m_selection);
}

void CVclipView::OnUpdateEditPaste(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(!m_clipboard.empty());
}
void CVclipView::OnEditPaste()
{
	YK_TRACED("OnEditPaste");
	DragCancel();
	SelectOperation(m_clipboard.paste(*GetDocument(), m_viewMode, m_pos, m_order, m_overlayMode));
}

void CVclipView::OnUpdateEditDelete(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canErase());
}
void CVclipView::OnEditDelete()
{
	m_manipulator->erase(*GetDocument());
	m_manipulator = std::auto_ptr<IManipulator>(new NullManipulator);
}

void CVclipView::OnUpdateEditSelectall(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditSelectall()
{
	UnselectOperation();
	m_selection = m_aviFile.frameSize();
	Invalidate();
}

void CVclipView::OnUpdateEditInsertimage(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditInsertimage()
{
	CFileDialog openDialog(TRUE, ".bmp", 0, OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST,
		/*1*/"Bitmap Files (*.bmp)|*.bmp|"
		/*2*/"GIF Files (*.gif)|*.gif|"
		/*3*/"JPEG Files (*.jpg)|*.jpg|"
		/*4*/"PNG Files (*.png)|*.png|"
		/*5*/"All Image Files (*.bmp;*.gif;*.jpg;*.png)|*.bmp;*.gif;*.jpg;*.png|"
		/*6*/"All Files (*.*)|*.*||", GetParent());
	openDialog.GetOFN().nFilterIndex = 5;

	if (IDOK == openDialog.DoModal())
	{
		SelectOperation(GetDocument()->add(
			the<FrameOperationFactory>().createFrameImage(*GetDocument(),
				m_pos, openDialog.GetFileName().GetString() ) ));
	}
}

void CVclipView::OnUpdateEditRebuildfromadjacentfields(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pos >= 1 && m_pos + 1 < m_aviFile.size());
}
void CVclipView::OnEditRebuildfromadjacentfields()
{
	SelectOperation(GetDocument()->add(
		the<FrameOperationFactory>().createFieldRebuild(*GetDocument(), m_pos) ));
}

void CVclipView::OnUpdateEditInterpolatemotion(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pos >= 1 && m_pos + 1 < m_aviFile.size());
}
void CVclipView::OnEditInterpolatemotion()
{
	SelectOperation(GetDocument()->add(
		the<FrameOperationFactory>().createMotionInterpolate(*GetDocument(), m_pos) ));
}

void CVclipView::OnUpdateEditSnapshotframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditSnapshotframe()
{
	CString defaultFilename;
	defaultFilename.Format("%d.bmp", m_pos);
	CFileDialog fd(FALSE, ".bmp", defaultFilename, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"Bitmap Files (*.bmp)|*.bmp|"
		"All Files (*.*)|*.*||", GetParent());
	if (IDOK != fd.DoModal()) return;

	PFrame frame = PrepareFrame(GetDC());

	std::ofstream ofs;
	ofs.exceptions(std::ios::failbit);
	ofs.open(fd.GetFileName(), std::ios_base::out | std::ios_base::binary);

	BITMAPFILEHEADER bfh;
	bfh.bfType = 0x4D42;
	bfh.bfSize = 
		sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER) +
		frame.header()->biSizeImage;
	bfh.bfReserved1 = bfh.bfReserved2 = 0;
	bfh.bfOffBits = 
		sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER);

	ofs.write(reinterpret_cast<const char*>(&bfh),           sizeof(bfh));
	ofs.write(reinterpret_cast<const char*>(frame.header()), frame.header()->biSize);
	ofs.write(reinterpret_cast<const char*>(frame.data()),   frame.header()->biSizeImage);
}

//#Edit

// CVclipView

IMPLEMENT_DYNCREATE(CVclipView, CView)

BEGIN_MESSAGE_MAP(CVclipView, CView)
	ON_COMMAND(ID_VIEW_OPENVIDEO, OnViewOpenvideo)
	ON_COMMAND(ID_VIEW_NEXTFRAME, OnViewNextframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_NEXTFRAME, OnUpdateViewNextframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PREVIOUSFRAME, OnUpdateViewPreviousframe)
	ON_COMMAND(ID_VIEW_PREVIOUSFRAME, OnViewPreviousframe)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateEditCopy)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_COMMAND(ID_DRAGCANCEL, OnDragcancel)
	ON_UPDATE_COMMAND_UI(ID_EDIT_DELETE, OnUpdateEditDelete)
	ON_COMMAND(ID_EDIT_DELETE, OnEditDelete)
	ON_COMMAND(ID_FILE_COMPILE, OnFileCompile)
	ON_COMMAND(ID_COMPILE_EVENCOORDINATESONLY, OnCompileEvencoordinatesonly)
	ON_UPDATE_COMMAND_UI(ID_COMPILE_EVENCOORDINATESONLY, OnUpdateCompileEvencoordinatesonly)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_FRAME, OnUpdateIndicatorFrame)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_FPS, OnUpdateIndicatorFps)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_MEDIATIME, OnUpdateIndicatorMediatime)
	ON_WM_CREATE()
	ON_WM_SIZE()
	ON_WM_HSCROLL()
	ON_COMMAND(ID_VIEW_GOTOFRAME, OnViewGotoframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_FRAMEMODE, OnUpdateViewFramemode)
	ON_UPDATE_COMMAND_UI(ID_VIEW_FIELDMODE, OnUpdateViewFieldmode)
	ON_COMMAND(ID_VIEW_FRAMEMODE, OnViewFramemode)
	ON_COMMAND(ID_VIEW_FIELDMODE, OnViewFieldmode)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECT_ALL, OnUpdateEditSelectall)
	ON_COMMAND(ID_EDIT_SELECT_ALL, OnEditSelectall)
	ON_UPDATE_COMMAND_UI(ID_VIEW_GOTOFRAME, OnUpdateViewGotoframe)
	ON_UPDATE_COMMAND_UI(ID_EDIT_INSERTIMAGE, OnUpdateEditInsertimage)
	ON_COMMAND(ID_EDIT_INSERTIMAGE, OnEditInsertimage)
	ON_UPDATE_COMMAND_UI(ID_EDIT_REBUILDFROMADJACENTFIELDS, OnUpdateEditRebuildfromadjacentfields)
	ON_COMMAND(ID_EDIT_REBUILDFROMADJACENTFIELDS, OnEditRebuildfromadjacentfields)
	ON_COMMAND(ID_EDIT_INTERPOLATEMOTION, OnEditInterpolatemotion)
	ON_UPDATE_COMMAND_UI(ID_EDIT_INTERPOLATEMOTION, OnUpdateEditInterpolatemotion)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_CHROMA, OnUpdateOverlaymodeChroma)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_LUMA, OnUpdateOverlaymodeLuma)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_OPAQUE, OnUpdateOverlaymodeOpaque)
	ON_COMMAND(ID_OVERLAYMODE_CHROMA, OnOverlaymodeChroma)
	ON_COMMAND(ID_OVERLAYMODE_LUMA, OnOverlaymodeLuma)
	ON_COMMAND(ID_OVERLAYMODE_OPAQUE, OnOverlaymodeOpaque)
	ON_WM_SETCURSOR()
	ON_WM_SETTINGCHANGE()
	ON_LBN_SELCHANGE(ID_FRAME_HISTORY, OnHistorySelChange)
	ON_UPDATE_COMMAND_UI(ID_Z_SENDTOBACK, OnUpdateZSendtoback)
	ON_UPDATE_COMMAND_UI(ID_Z_SENDFURTHER, OnUpdateZSendfurther)
	ON_UPDATE_COMMAND_UI(ID_Z_BRINGNEARER, OnUpdateZBringnearer)
	ON_UPDATE_COMMAND_UI(ID_Z_BRINGTOFRONT, OnUpdateZBringtofront)
	ON_COMMAND(ID_Z_SENDTOBACK, OnZSendtoback)
	ON_COMMAND(ID_Z_SENDFURTHER, OnZSendfurther)
	ON_COMMAND(ID_Z_BRINGNEARER, OnZBringnearer)
	ON_COMMAND(ID_Z_BRINGTOFRONT, OnZBringtofront)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVEDOWN, OnUpdateEditMovedown)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVELEFT, OnUpdateEditMoveleft)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVERIGHT, OnUpdateEditMoveright)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVEUP, OnUpdateEditMoveup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZELEFT, OnUpdateEditSizeleft)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZERIGHT, OnUpdateEditSizeright)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZEUP, OnUpdateEditSizeup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZEDOWN, OnUpdateEditSizedown)
	ON_COMMAND(ID_EDIT_MOVEDOWN, OnEditMovedown)
	ON_COMMAND(ID_EDIT_MOVELEFT, OnEditMoveleft)
	ON_COMMAND(ID_EDIT_MOVERIGHT, OnEditMoveright)
	ON_COMMAND(ID_EDIT_MOVEUP, OnEditMoveup)
	ON_COMMAND(ID_EDIT_SIZEDOWN, OnEditSizedown)
	ON_COMMAND(ID_EDIT_SIZELEFT, OnEditSizeleft)
	ON_COMMAND(ID_EDIT_SIZERIGHT, OnEditSizeright)
	ON_COMMAND(ID_EDIT_SIZEUP, OnEditSizeup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECTNEXT, OnUpdateEditSelectnext)
	ON_COMMAND(ID_EDIT_SELECTNEXT, OnEditSelectnext)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECTPREVIOUS, OnUpdateEditSelectprevious)
	ON_COMMAND(ID_EDIT_SELECTPREVIOUS, OnEditSelectprevious)
	ON_UPDATE_COMMAND_UI(ID_VIEW_NEXTMODIFIED, OnUpdateViewNextmodified)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PREVIOUSMODIFIED, OnUpdateViewPreviousmodified)
	ON_COMMAND(ID_VIEW_NEXTMODIFIED, OnViewNextmodified)
	ON_COMMAND(ID_VIEW_PREVIOUSMODIFIED, OnViewPreviousmodified)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SNAPSHOTFRAME, OnUpdateEditSnapshotframe)
	ON_COMMAND(ID_EDIT_SNAPSHOTFRAME, OnEditSnapshotframe)
	ON_COMMAND(ID_VIEW_INVERT, OnViewInvert)
	ON_UPDATE_COMMAND_UI(ID_VIEW_INVERT, OnUpdateViewInvert)
	ON_UPDATE_COMMAND_UI(ID_VIEW_DEINTERLACE, OnUpdateViewDeinterlace)
	ON_COMMAND(ID_VIEW_DEINTERLACE, OnViewDeinterlace)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ACCURATERENDERING, OnUpdateViewAccuraterendering)
	ON_COMMAND(ID_VIEW_ACCURATERENDERING, OnViewAccuraterendering)
END_MESSAGE_MAP()

//:View maintenance

CVclipView::CVclipView()
: m_aviFile(),
	m_drawDib(),
	m_pos(0), m_order(0), m_viewMode(&frameMode), m_overlayMode(the<Overlays>()["Blend"]),
	m_dragState(new NullDrag), m_manipulator(new SelectionManipulator(m_viewMode, m_selection, CRect())),
	m_selection(0, 0, 0, 0),
	m_clipboard(),
	m_evenCoords(true), m_accurateRendering(false),
	m_invert(&identity), m_deinterlace(&identity)
{
}

CVclipView::~CVclipView()
{
}

BOOL CVclipView::PreCreateWindow(CREATESTRUCT& cs)
{
	BOOL result = CView::PreCreateWindow(cs);
	cs.dwExStyle |= WS_EX_COMPOSITED;
	return result;
}

int CVclipView::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;

	m_frameTrackBar.Create(WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_BOTTOM | TBS_HORZ,
		CRect(0, lpCreateStruct->cy - 20, lpCreateStruct->cx, lpCreateStruct->cy),
		this, ID_FRAME_TRACKBAR);
	m_frameTrackBar.SetRange(0, 0);

	m_historyList.Create(WS_CHILD | WS_VISIBLE | LBS_NOTIFY,
		CRect(0, lpCreateStruct->cy - 200, lpCreateStruct->cx, lpCreateStruct->cy - 20),
		this, ID_FRAME_HISTORY);
	UpdateHistoryFont();
	
	return 0;
}

void CVclipView::OnSize(UINT nType, int cx, int cy)
{
	CView::OnSize(nType, cx, cy);

	m_frameTrackBar.MoveWindow(0, cy - 20, cx, 20);
	m_historyList.MoveWindow(0, m_aviFile.frameSize().Height(), cx, cy - 20 - m_aviFile.frameSize().Height());
}

BOOL CVclipView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
	CPoint viewPoint;
	GetCursorPos(&viewPoint);
	ScreenToClient(&viewPoint);

	if (!m_viewMode->viewFromFrame(m_aviFile.frameSize()).PtInRect(viewPoint))
		return CView::OnSetCursor(pWnd, nHitTest, message);

	CPoint framePoint = m_viewMode->frameFromView(viewPoint);

	if (m_manipulator->setCursor(framePoint)) return TRUE;

	if (GetAsyncKeyState(VK_CONTROL))
	{
		SetCursor(LoadCursor(0, IDC_CROSS));
		return TRUE;
	}

	PFrameOperation op;
	GetDocument()->operations(m_pos).for_each(
		HitTest(m_viewMode, m_order, framePoint, op) );
	if (op)
	{
		SetCursor(LoadCursor(0, IDC_ARROW));
		return TRUE;
	}
	else
	{
		SetCursor(LoadCursor(0, IDC_CROSS));
		return TRUE;
	}
}

//#View maintenance

//:Diagnostics

#ifdef _DEBUG
void CVclipView::AssertValid() const
{
	CView::AssertValid();
}

void CVclipView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CVclipDoc* CVclipView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CVclipDoc)));
	return (CVclipDoc*)m_pDocument;
}
#endif //_DEBUG

//#Diagnostics

//:Video file

void CVclipView::OpenVideo(LPCTSTR lpszFilename)
{
	m_aviFile.open(lpszFilename);

	m_frameTrackBar.SetRange(0, m_aviFile.size());
	m_frameTrackBar.SetPageSize(m_aviFile.size() / 100);

	CRect rect;
	GetClientRect(rect);
	m_historyList.MoveWindow(0, m_aviFile.height(), rect.right, rect.bottom - 20 - m_aviFile.height());

	SetFrame(0);
}

void CVclipView::CloseVideo()
{
	m_aviFile.close();

	m_frameTrackBar.SetRange(0, 0);
	m_frameTrackBar.SetPageSize(0);

	CRect rect;
	GetClientRect(rect);
	m_historyList.MoveWindow(0, 0, rect.right, rect.bottom - 20);

	SetFrame(0);
}

void CVclipView::OnViewOpenvideo()
{
	CString context = "opening video";
	try
	{
		CFileDialog openDialog(TRUE, ".avi", 0,
			OFN_ENABLESIZING | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST,
			"Video files (*.avi, *.avs)|*.avi;*.avs|All files (*.*)|*.*||",
			this->GetParent());
		if (IDOK != openDialog.DoModal()) return;

		context.Format("getting full path name of %s", openDialog.GetFileName());
		TCHAR buf[MAX_PATH];
		LPTSTR filePart;
		ASSERT_NE(GetFullPathName(openDialog.GetFileName(), MAX_PATH, buf, &filePart), FALSE, "GetFullPathName");

		context.Format("opening video file %s", buf);
		GetDocument()->setSource(buf);
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//#Video file

//:Frame seeking

void CVclipView::SetFrame(LONG frame)
{
	if (!GetDocument()->needFrame(m_pos))
	{
		m_aviFile.erase(m_pos);
	}
	
	DragCancel();
	UnselectOperation();
	
	m_pos = frame;
	m_frameTrackBar.SetPos(m_pos);
	UpdateHistory();

	Invalidate();
}

void CVclipView::OnUpdateViewNextframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_viewMode->hasNext(m_pos, m_order, m_aviFile.size()));
}
void CVclipView::OnViewNextframe()
{
	SetFrame(m_viewMode->next(m_pos, m_order));
}

void CVclipView::OnUpdateViewPreviousframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_viewMode->hasPrevious(m_pos, m_order));
}
void CVclipView::OnViewPreviousframe()
{
	SetFrame(m_viewMode->previous(m_pos, m_order));
}

void CVclipView::OnUpdateViewPreviousmodified(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(GetDocument()->operations(m_pos).first != GetDocument()->begin());
}
void CVclipView::OnViewPreviousmodified()
{
	SetFrame((*(GetDocument()->operations(m_pos).first - 1))->targetFrame());
}

void CVclipView::OnUpdateViewNextmodified(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(GetDocument()->operations(m_pos).second != GetDocument()->end());
}
void CVclipView::OnViewNextmodified()
{
	SetFrame((*GetDocument()->operations(m_pos).second)->targetFrame());
}

void CVclipView::OnUpdateViewGotoframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnViewGotoframe()
{
	CGotoFrame gotoDialog;
	gotoDialog.SetFrame(m_pos);
	gotoDialog.SetFps(m_aviFile.millifps());
	gotoDialog.SetFrames(m_aviFile.size());
	if (IDOK == gotoDialog.DoModal())
	{
		SetFrame(gotoDialog.GetFrame());
	}
}

void CVclipView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
// RATIONALE: (not using nPos)
// Trackbar sends a WM_SCROLL message.
// nPos is contained in HIWORD(wParam), which already limits positions to 16 bits.
// MFC, additionally, reinterprets it as signed short, which yields negative values
// if nPos > 32767.
{
	if (reinterpret_cast<CSliderCtrl*>(pScrollBar) == &m_frameTrackBar)
	{
		switch (nSBCode)
		{
		case TB_LINEUP:   case TB_PAGEUP:   case TB_TOP:
		case TB_LINEDOWN: case TB_PAGEDOWN: case TB_BOTTOM:
		case TB_THUMBPOSITION:
		case TB_THUMBTRACK:
		case TB_ENDTRACK:
			SetFrame(m_frameTrackBar.GetPos());
			return;
		default:
			return;
		}
	}
	else
	{
		CView::OnHScroll(nSBCode, nPos, pScrollBar);
	}
}

//#Frame seeking

//:History box

void CVclipView::UpdateHistoryFont()
{
	LOGFONT lf;
	ASSERT_NE(SystemParametersInfo(SPI_GETICONTITLELOGFONT, sizeof(lf), &lf, FALSE), 
		FALSE, "SystemParametersInfo");
	m_font.DeleteObject();
	ASSERT_NE(m_font.CreateFontIndirect(&lf),
		FALSE, "CreateFontIndirect");
	m_historyList.SetFont(&m_font);
}

void CVclipView::OnSettingChange(UINT uFlags, LPCTSTR lpszSection)
{
	CView::OnSettingChange(uFlags, lpszSection);
	if (uFlags == 0 || uFlags == SPI_SETICONTITLELOGFONT || uFlags == SPI_GETICONTITLELOGFONT)
	{
		UpdateHistoryFont();
	}
}

class AddOperationToHistoryBox : public std::unary_function<PFrameOperation, void>
{
private:
	CListBox* m_pBox;
public:
	explicit AddOperationToHistoryBox(CListBox& box)
		: m_pBox(&box)
	{}

	void operator()(PFrameOperation op) const
	{
		std::ostringstream oss;
		oss << op;
		m_pBox->SetItemDataPtr(
			m_pBox->AddString(oss.str().c_str()), 
			op.get());
	}
};

void CVclipView::UpdateHistory()
{
	m_historyList.ResetContent();
	CVclipDoc::range operations = GetDocument()->operations(m_pos);
	std::for_each(operations.first, operations.second,
		AddOperationToHistoryBox(m_historyList));
}

void CVclipView::OnHistorySelChange()
{
	int sel = m_historyList.GetCurSel();
	if (sel < 0) return;
	SelectOperation(PFrameOperation(
		static_cast<FrameOperation*>(m_historyList.GetItemDataPtr(sel))));
}

//#History box

//:Overlay mode

class UpdateOverlayMode : public IOperationVisitor
{
private:
	HOverlay* m_pMode;
public:
	explicit UpdateOverlayMode(HOverlay* pMode) : m_pMode(pMode) {}

	void operator()(FramePatch& patch) { *m_pMode = patch.mode(); }
	void operator()(FieldPatch& patch) { *m_pMode = patch.mode(); }
	void operator()(FrameImage& /*image*/) {}
	void operator()(FieldRebuild& /*rebuild*/) {}
	void operator()(MotionInterpolate& /*inter*/) {}
};

class SetOverlayMode : public IOperationVisitor
{
private:
	HOverlay m_mode;
public:
	explicit SetOverlayMode(HOverlay mode) : m_mode(mode) {}

	void operator()(FramePatch& patch) { patch.setMode(m_mode); }
	void operator()(FieldPatch& patch) { patch.setMode(m_mode); }
	void operator()(FrameImage& /*image*/) {}
	void operator()(FieldRebuild& /*rebuild*/) {}
	void operator()(MotionInterpolate& /*inter*/) {}
};

void CVclipView::UpdateOverlaymode(CCmdUI* pCmdUI, HOverlay mode) const
{
	pCmdUI->Enable(m_aviFile.size());
	pCmdUI->SetRadio(m_overlayMode == mode);
}
void CVclipView::SetOverlaymode(HOverlay mode)
{
	m_manipulator->accept(SetOverlayMode(mode));
	m_overlayMode = mode;
}

void CVclipView::OnUpdateOverlaymodeOpaque(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Blend"]);
}
void CVclipView::OnOverlaymodeOpaque()
{
	SetOverlaymode(the<Overlays>()["Blend"]);
}

void CVclipView::OnUpdateOverlaymodeChroma(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Chroma"]);
}
void CVclipView::OnOverlaymodeChroma()
{
	SetOverlaymode(the<Overlays>()["Chroma"]);
}

void CVclipView::OnUpdateOverlaymodeLuma(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Luma"]);
}
void CVclipView::OnOverlaymodeLuma()
{
	SetOverlaymode(the<Overlays>()["Luma"]);
}

//#Overlay mode

//:Z-order

void CVclipView::OnUpdateZSendtoback(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation());
}
void CVclipView::OnZSendtoback()
{
	GetDocument()->sendToBack(m_manipulator->operation());
}

void CVclipView::OnUpdateZSendfurther(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation() && 
		GetDocument()->canSendFurther(m_manipulator->operation()));
}
void CVclipView::OnZSendfurther()
{
	GetDocument()->sendFurther(m_manipulator->operation());
}

void CVclipView::OnUpdateZBringnearer(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation() &&
		GetDocument()->canBringNearer(m_manipulator->operation()));
}
void CVclipView::OnZBringnearer()
{
	GetDocument()->bringNearer(m_manipulator->operation());
}

void CVclipView::OnUpdateZBringtofront(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation());
}
void CVclipView::OnZBringtofront()
{
	GetDocument()->bringToFront(m_manipulator->operation());
}

//#Z-order

//:Compilation

void CVclipView::OnUpdateCompileEvencoordinatesonly(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_evenCoords);
}
void CVclipView::OnCompileEvencoordinatesonly()
{
	m_evenCoords = !m_evenCoords;
}

void CVclipView::OnFileCompile()
{
	YK_TRACED("OnFileCompile");

	CFileDialog saveDialog(FALSE, ".avs", 0, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"AviSynth Files (*.avs)|*.avs|All Files (*.*)|*.*||", GetParent());
	if (IDOK == saveDialog.DoModal())
	{
		GetDocument()->compile(saveDialog.GetFileName(), m_evenCoords, 
			m_aviFile.size(), m_aviFile.frameSize());
	}
}

//#Compilation

//:Document

void CVclipView::OnUpdate(CView* /*pSender*/, LPARAM lHint, CObject* /*pHint*/)
{
	CString context = "updating";
	try
	{
		if (lHint < 0)
		{
			// source changed
			CloseVideo();

			if (!GetDocument()->source().empty())
			{
				context.Format("opening video file %s", GetDocument()->source().c_str());
				OpenVideo(GetDocument()->source().c_str());
			}
		}
		if (lHint == m_pos || lHint == -2)
		{
			UpdateHistory();
			SelectOperation(m_manipulator->operation());
			Invalidate();
		}
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//:Status bar

void CVclipView::OnUpdateIndicatorFrame(CCmdUI *pCmdUI)
{
	pCmdUI->Enable();
	pCmdUI->SetText(m_viewMode->formatPosition(m_pos, m_order, m_aviFile.size()));
}

void CVclipView::OnUpdateIndicatorMediatime(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size() != 0);
	if (m_aviFile.size())
		pCmdUI->SetText(formatMediaTime(mediaTimeFromFrame(m_pos, m_aviFile.millifps())));
}

void CVclipView::OnUpdateIndicatorFps(CCmdUI *pCmdUI)
{
	pCmdUI->Enable();
	CString text;
	text.Format("%d.%03d fps", m_aviFile.millifps() / 1000, m_aviFile.millifps() % 1000);
	pCmdUI->SetText(text);
}

//#Status bar

/////
